/******************************************************************

        MUSCLE SmartCard Development ( http://www.linuxnet.com )
            Title  : challenge.c
            Package: RSA Smartcard PAM module
            Author : David Corcoran
            Date   : 4/12/00
            Purpose: This extracts and encrypts using RSA public keys
	    and tests a challenge response.

********************************************************************/

#include <includes.h>
#include <ssh.h>
#include <gmp.h>
#include <xmalloc.h>
#include <rsa.h>
#include <stdio.h>
#include <winscard.h>
#include <sys/timeb.h>
#include <unistd.h>

extern UCHAR ACCESS_CHV0_VAL[8];
extern DWORD ACCESS_CHV0_SIZE;
extern UCHAR ACCESS_AUT0_KEY[8];
extern DWORD ACCESS_AUT0_SIZE;

void PMSCGenerateRandomNonce( PUCHAR pucBuffer, DWORD dwLength );


void PMSCGenerateRandomNonce( PUCHAR pucBuffer, DWORD dwLength ) {

  UCHAR ucRByte;
  int i;
  int iRandom;

  /* Start with a decent seed and then feed the random
     as the seed - use the Epoch and pid.
  */

  iRandom  = time(0) + getpid();

  for ( i=0; i < dwLength; i++ ) {
    srand( iRandom + i );
    ucRByte = (1 + (int)(255.0*rand()/(RAND_MAX+1.0)));
    pucBuffer[i] = ucRByte;
  }

  return;
}


int main( int argc, char **argv ) {


  SCARDCONTEXT hContext;
  SCARDHANDLE hCard;
  UCHAR pucRandom[200], pucCipherText[200], pucPlainText[200];
  RSAPublicKey Kpub;
  DWORD dwSize;
  DWORD dwProtocol, dwReaders;
  LPCSTR mszGroups;
  LPSTR mszReaders;
  SCARD_READERSTATE_A rgReaderStates;
  LONG rv;
  int i, p, iReader, fd, result;
  int iList[16];
  char sUser[100];

  sprintf(sUser, "%s%s", getenv("HOME"), "/.muscle/public.key");

  /* Get the public key and generate the nonce */
  RSAKpubGetKey( &Kpub, sUser );  
  PMSCGenerateRandomNonce( pucRandom, 128 );
	
  /* Padding for correct crypto operation  */
  pucRandom[0] = 0x80; pucRandom[127] = 0x01;

  /* Encrypt the nonce with the public key */
  dwSize = 128;
  memset( pucCipherText, 0x00, 128 );
  RSAKpubEncrypt( &Kpub, pucRandom, pucCipherText, &dwSize );

  /* Connect to the smartcard subsystem */
  SCardEstablishContext(SCARD_SCOPE_SYSTEM, 0, 0, &hContext);

  mszGroups = 0;
  SCardListReaders( hContext, mszGroups, 0, &dwReaders );
  mszReaders = (char *)malloc(sizeof(char)*dwReaders);
  SCardListReaders( hContext, mszGroups, mszReaders, &dwReaders );

  /* Have to understand the multi-string here */
  p = 0;
  for ( i=0; i < dwReaders - 1; i++ ) {
    ++p;
    printf("Reader %02d: %s\n", p, &mszReaders[i]);
    iList[p] = i;
    while ( mszReaders[++i] != 0 );
  }

  do {
    printf("Enter the desired reader number : " );
    scanf("%d", &iReader);
    printf("\n");

    if ( iReader > p || iReader <= 0 ) {
      printf("Invalid Value - try again\n");
    }
  } while ( iReader > p || iReader <= 0 );

  rgReaderStates.szReader       = &mszReaders[iList[iReader]];
  rgReaderStates.dwCurrentState = SCARD_STATE_EMPTY;

  /* Wait for a card to be inserted */
  SCardGetStatusChange( hContext, INFINITE, &rgReaderStates, 1 );

  SCardConnect( hContext, &mszReaders[iList[iReader]], SCARD_SHARE_SHARED, 
		SCARD_PROTOCOL_T0, &hCard, &dwProtocol );

  /* Verify the transport key and the chv key
     Note - the chv will also be used in future releases 
  */
  rv = SCSVerifyKey( hCard, ACCESS_AUT0_SIZE, ACCESS_AUT0_KEY );
  rv = SCSVerifyChv( hCard, ACCESS_CHV0_SIZE, ACCESS_CHV0_VAL );

  /* Sign the crypted nonce using the card's private key */
  dwSize = 128;
  memset( pucPlainText, 0x00, sizeof(pucPlainText) );
  rv = SCSRSASignData( hCard, pucCipherText, pucPlainText, &dwSize );

  /* Disconnect from the smartcard subsystem */
  rv = SCardDisconnect( hCard, SCARD_LEAVE_CARD );
  rv = SCardReleaseContext( hContext );

  result = memcmp( pucRandom, pucPlainText, 128 );

  if ( result == 0 ) {
    printf("Challenge completed successfully !\n");
  } else {
    printf("Challenge failed miserably !\n");
  }

  return 0;
}
