/******************************************************************

        MUSCLE SmartCard Development ( http://www.linuxnet.com )
            Title  : csgenkey.c
            Package: RSA Smartcard PAM module
            Author : David Corcoran <corcoran@linuxnet.com>
            Author : Nam Itoi       <itoi@umich.edu>
            Date   : 4/11/00
            Purpose: This generates and writes RSA keys to 
	             smartcards.

********************************************************************/

#include <includes.h>
#include <ssh.h>
#include <gmp.h>
#include <xmalloc.h>
#include <rsa.h>
#include <stdio.h>
#include <winscard.h>


int main(int argc, char **argv) {

  SCARDCONTEXT hContext;
  SCARDHANDLE  hCard;
  RSAPublicKey  Kpub, KpubB;
  RSAPrivateKey Kpriv;
  RandomState state;
  unsigned char buf[1024];
  struct passwd *pw;
  DWORD dwProtocol, dwReaders;
  LPCSTR mszGroups;
  LPSTR mszReaders;
  SCARD_READERSTATE_A rgReaderStates;
  UCHAR GCRTKeyVal[1000];
  LONG rv;
  int i, p, iReader, fd;
  int iList[16];
  char sUser[100];

  pw = getpwuid(getuid());

  /* Initialize random number generator.  This may take a while if the
     user has no seed file, so display a message to the user. */
  printf("Initializing random number generator...\n");
  snprintf(buf, sizeof(buf), "%s/%s", pw->pw_dir, SSH_CLIENT_SEEDFILE);   
  random_initialize(&state, geteuid(), buf);

  /* Save random seed so we don't need to do all that time-consuming
     environmental noise collection the next time. */
  random_save(&state, geteuid(), buf);

  rsa_generate_key(&Kpriv, &Kpub, &state, 1024);

  printf("\n ---------- Connecting to smartcard subsystem ----------\n\n");

  rv = SCardEstablishContext(SCARD_SCOPE_SYSTEM, 0, 0, &hContext);

  if ( rv != SCARD_S_SUCCESS ) exit(1);

  mszGroups = 0;
  SCardListReaders( hContext, mszGroups, 0, &dwReaders );
  mszReaders = (char *)malloc(sizeof(char)*dwReaders);
  SCardListReaders( hContext, mszGroups, mszReaders, &dwReaders );

  /* Have to understand the multi-string here */
  p = 0;
  for ( i=0; i < dwReaders - 1; i++ ) {
    ++p;
    printf("Reader %02d: %s\n", p, &mszReaders[i]);
    iList[p] = i;
    while ( mszReaders[++i] != 0 );
  }

  do {
    printf("Enter the desired reader number : " );
    scanf("%d", &iReader);
    printf("\n");

    if ( iReader > p || iReader <= 0 ) {
      printf("Invalid Value - try again\n");
    }
  } while ( iReader > p || iReader <= 0 );

  rgReaderStates.szReader       = &mszReaders[iList[iReader]];
  rgReaderStates.dwCurrentState = SCARD_STATE_EMPTY;

  printf("Please insert your smartcard: ");
  fflush(stdout);

  SCardGetStatusChange( hContext, 0, &rgReaderStates, 1 );

  printf("card inserted\n");

  SCardConnect( hContext, &mszReaders[iList[iReader]], SCARD_SHARE_SHARED, 
		SCARD_PROTOCOL_T0, &hCard, &dwProtocol );

  /* Formats the Private key in bytes then loads it to the card */
  SCSFormatRsaKPriv( &Kpriv, GCRTKeyVal );
  SCSLoadRsaKPriv ( hCard, GCRTKeyVal, 64*6 );

  sprintf(sUser, "%s%s", getenv("HOME"), "/.muscle");
  mkdir( sUser, S_IRWXU | S_IROTH );

  sprintf(sUser, "%s%s", getenv("HOME"), "/.muscle/public.key");

  fd = open(sUser, O_CREAT | O_WRONLY, S_IRWXU | S_IROTH ); 

  if ( fd <= 0 ) {
    printf("Unable to open/create public key ~/.muscle/public.key\n");
    printf("Does the directory ~/.muscle exist ????\n");
    SCardDisconnect( hCard, SCARD_LEAVE_CARD );
    SCardReleaseContext( hContext );
    exit(1);
  }

  printf("Writing RSA Public Key: ~/.muscle/public.key\n\n");
  mp_linearize_msb_first( buf, 4, &Kpub.e );
  write( fd, buf, 4 );
  mp_linearize_msb_first( buf, 128, &Kpub.n );
  write( fd, buf, 128 );
  close(fd);

  SCardDisconnect( hCard, SCARD_LEAVE_CARD );
  SCardReleaseContext( hContext );

  printf(" --------------- Finished Key Generation ---------------\n\n");

}
